from __future__ import annotations

import atexit
import functools
import logging
import multiprocessing
import os
import shutil
import signal
import socket
import subprocess
import tempfile
import time
import webbrowser
from collections import namedtuple
from contextlib import closing
from enum import Enum, auto
from io import BufferedWriter
from os import path
from typing import Dict, List, Optional, Set

import docker
import grpc
import psutil
import requests
import toml
from urllib3.exceptions import MaxRetryError

import vega_sim.api.governance as gov
import vega_sim.grpc.client as vac
from vega_sim import vega_bin_path, vega_home_path
from vega_sim.service import VegaService
from vega_sim.tools.load_binaries import download_binaries
from vega_sim.tools.retry import retry
from vega_sim.wallet.base import DEFAULT_WALLET_NAME, Wallet
from vega_sim.wallet.slim_wallet import SlimWallet
from vega_sim.wallet.vega_wallet import VegaWallet

logger = logging.getLogger(__name__)

PortUpdateConfig = namedtuple(
    "PortUpdateConfig", ["file_path", "config_path", "key", "val_func"]
)


class Ports(Enum):
    DATA_NODE_GRPC = auto()
    DATA_NODE_REST = auto()
    DATA_NODE_POSTGRES = auto()
    FAUCET = auto()
    WALLET = auto()
    VEGA_NODE = auto()
    CORE_GRPC = auto()
    CORE_REST = auto()
    BROKER = auto()
    METRICS = auto()
    PPROF = auto()
    CONSOLE = auto()


PORT_UPDATERS = {
    Ports.DATA_NODE_GRPC: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["API"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Gateway", "Node"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "GRPC"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
    ],
    Ports.DATA_NODE_REST: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Gateway"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "REST"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "GraphQL"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
    ],
    Ports.DATA_NODE_POSTGRES: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["SQLStore", "ConnectionConfig"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.FAUCET: [
        PortUpdateConfig(
            ("config", "faucet", "config.toml"), [], "Port", lambda port: port
        ),
    ],
    Ports.WALLET: [
        PortUpdateConfig(
            ("config", "wallet-service", "config.toml"),
            ["Server"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            [],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.VEGA_NODE: [
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Blockchain", "Null"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.CORE_GRPC: [
        PortUpdateConfig(
            ("config", "faucet", "config.toml"),
            ["Node"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["API"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["API"],
            "CoreNodeGRPCPort",
            lambda port: port,
        ),
    ],
    Ports.CORE_REST: [
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["API", "REST"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.BROKER: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Broker", "SocketConfig"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Broker", "Socket"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.METRICS: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Metrics"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Metrics"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.PPROF: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Pprof"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Pprof"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.CONSOLE: [
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["Console"],
            "LocalPort",
            lambda port: port,
        ),
    ],
}


class VegaStartupTimeoutError(Exception):
    pass


class ServiceNotStartedError(Exception):
    pass


class SocketNotFoundError(Exception):
    pass


def find_free_port(existing_set: Optional[Set[int]] = None):
    ret_sock = 0
    existing_set = (
        existing_set.union(set([ret_sock]))
        if existing_set is not None
        else set([ret_sock])
    )

    num_tries = 0
    while ret_sock in existing_set:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            ret_sock = s.getsockname()[1]

        num_tries += 1
        if num_tries >= 100:
            # Arbitrary high number. If we try 100 times and fail to find
            # a port it seems reasonable to give up
            raise SocketNotFoundError("Failed finding a free socket")

    return ret_sock


def _terminate_proc(
    proc: subprocess.Popen[bytes], out_file: BufferedWriter, err_file: BufferedWriter
) -> None:
    proc.terminate()
    out_file.close()
    err_file.close()


def _popen_process(
    popen_args: List[str],
    dir_root: str,
    log_name: str,
    env: Optional[Dict[str, str]] = None,
) -> subprocess.Popen[bytes]:
    out = open(path.join(dir_root, f"{log_name}.out"), "wb")
    err = open(path.join(dir_root, f"{log_name}.err"), "wb")
    sub_proc = subprocess.Popen(
        popen_args, stdout=out, stderr=err, env=env, close_fds=True
    )

    atexit.register(functools.partial(_terminate_proc, sub_proc, out, err))
    return sub_proc


def _update_node_config(
    vega_home: str,
    port_config: Dict[Ports, int],
    transactions_per_block: int = 1,
    block_duration: str = "1s",
    use_docker_postgres: bool = False,
) -> None:
    config_path = path.join(vega_home, "config", "node", "config.toml")
    config_toml = toml.load(config_path)
    config_toml["Blockchain"]["Null"]["GenesisFile"] = path.join(
        vega_home, "genesis.json"
    )
    config_toml["Blockchain"]["Null"]["BlockDuration"] = block_duration
    config_toml["Blockchain"]["Null"]["TransactionsPerBlock"] = transactions_per_block

    existing_ports = set(port_config.values())
    for port in Ports:
        if port in port_config:
            continue
        new_port = find_free_port(existing_ports)
        existing_ports.add(new_port)
        port_config[port] = new_port

    with open(config_path, "w") as f:
        toml.dump(config_toml, f)

    for port_key, update_configs in PORT_UPDATERS.items():
        for config in update_configs:
            file_path = path.join(vega_home, *config.file_path)
            config_toml = toml.load(file_path)
            elem = config_toml
            for k in config.config_path:
                elem = elem[k]
            elem[config.key] = config.val_func(port_config[port_key])

            if port_key == Ports.DATA_NODE_POSTGRES:
                config_toml["SQLStore"]["UseEmbedded"] = not use_docker_postgres

            with open(file_path, "w") as f:
                toml.dump(config_toml, f)


def manage_vega_processes(
    child_conn: multiprocessing.Pipe,
    vega_path: str,
    data_node_path: str,
    vega_wallet_path: str,
    vega_console_path: Optional[str] = None,
    run_with_console: bool = False,
    port_config: Optional[Dict[Ports, int]] = None,
    transactions_per_block: int = 1,
    block_duration: str = "1s",
    run_wallet: bool = False,
    retain_log_files: bool = False,
    log_dir: Optional[str] = None,
    replay_from_path: Optional[str] = None,
    store_transactions: bool = True,
) -> None:
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s: %(message)s",
    )
    port_config = port_config if port_config is not None else {}

    try:
        docker_client = docker.from_env()
        use_docker_postgres = True
    except:
        use_docker_postgres = False

    # Explicitly not using context here so that crashed logs are retained
    tmp_vega_dir = tempfile.mkdtemp(prefix="vega-sim-") if log_dir is None else log_dir
    logger.info(f"Running NullChain from vegahome of {tmp_vega_dir}")

    if port_config.get(Ports.CONSOLE):
        logger.info(f"Launching Console at port {port_config.get(Ports.CONSOLE)}")
    if port_config.get(Ports.DATA_NODE_REST):
        logger.info(
            "Launching Datanode REST + GRAPHQL at port"
            f" {port_config.get(Ports.DATA_NODE_REST)}"
        )
    if port_config.get(Ports.DATA_NODE_GRPC):
        logger.info(
            f"Launching Datanode GRPC at port {port_config.get(Ports.DATA_NODE_GRPC)}"
        )
    if port_config.get(Ports.CORE_REST):
        logger.info(f"Launching Core REST at port {port_config.get(Ports.CORE_REST)}")

    if port_config.get(Ports.CORE_GRPC):
        logger.info(f"Launching Core GRPC at port {port_config.get(Ports.CORE_GRPC)}")

    dest_dir = f"{tmp_vega_dir}/vegahome"
    shutil.copytree(vega_home_path, dest_dir)
    for dirpath, _, filenames in os.walk(dest_dir):
        os.utime(dirpath, None)
        for file in filenames:
            os.utime(os.path.join(dirpath, file), None)

    tmp_vega_home = tmp_vega_dir + "/vegahome"
    _update_node_config(
        tmp_vega_home,
        port_config=port_config,
        transactions_per_block=transactions_per_block,
        block_duration=block_duration,
        use_docker_postgres=use_docker_postgres,
    )

    if use_docker_postgres:
        data_node_docker_volume = docker_client.volumes.create()
        data_node_container = docker_client.containers.run(
            "timescale/timescaledb:2.8.0-pg14",
            detach=True,
            ports={5432: port_config[Ports.DATA_NODE_POSTGRES]},
            volumes=[f"{data_node_docker_volume.name}:/var/lib/postgresql/data"],
            environment={
                "POSTGRES_USER": "vega",
                "POSTGRES_PASSWORD": "vega",
                "POSTGRES_DB": "vega",
            },
            remove=True,
        )

    dataNodeProcess = _popen_process(
        [
            data_node_path,
            "start",
            "--home=" + tmp_vega_home,
            "--chainID=CUSTOM",
        ],
        dir_root=tmp_vega_dir,
        log_name="data_node",
    )

    vegaFaucetProcess = _popen_process(
        [
            vega_path,
            "faucet",
            "run",
            "--passphrase-file=" + tmp_vega_home + "/passphrase-file",
            "--home=" + tmp_vega_home,
        ],
        dir_root=tmp_vega_dir,
        log_name="faucet",
    )

    vega_args = [
        vega_path,
        "start",
        "--nodewallet-passphrase-file=" + tmp_vega_home + "/passphrase-file",
        "--home=" + tmp_vega_home,
    ]

    if store_transactions:
        replay_file = (
            replay_from_path
            if replay_from_path is not None
            else tmp_vega_home + "/replay"
        )
        vega_args.extend(
            [
                f"--blockchain.nullchain.replay-file={replay_file}",
                "--blockchain.nullchain.record",
            ]
        )
    if replay_from_path is not None:
        vega_args.extend(
            [
                f"--blockchain.nullchain.replay-file={replay_from_path}",
                "--blockchain.nullchain.replay",
            ]
        )

    vegaNodeProcess = _popen_process(
        vega_args,
        dir_root=tmp_vega_dir,
        log_name="node",
    )

    processes = {
        "data-node": dataNodeProcess,
        "faucet": vegaFaucetProcess,
        "vega": vegaNodeProcess,
    }

    if run_wallet:
        for _ in range(3000):
            try:
                requests.get(
                    f"http://localhost:{port_config.get(Ports.DATA_NODE_REST)}/time"
                ).raise_for_status()
                requests.get(
                    f"http://localhost:{port_config.get(Ports.CORE_REST)}/blockchain/height"
                ).raise_for_status()
                break
            except (
                MaxRetryError,
                requests.exceptions.ConnectionError,
                requests.exceptions.HTTPError,
            ):
                time.sleep(0.1)

        subprocess.run(
            [
                vega_wallet_path,
                "api-token",
                "init",
                f"--home={tmp_vega_home}",
                f"--passphrase-file={tmp_vega_home}/passphrase-file",
            ],
            capture_output=True,
        )

        subprocess.run(
            [
                vega_wallet_path,
                "create",
                "--wallet",
                DEFAULT_WALLET_NAME,
                "--home",
                tmp_vega_home,
                "--passphrase-file",
                tmp_vega_home + "/passphrase-file",
                "--output",
                "json",
            ],
            capture_output=True,
        )

        subprocess.run(
            [
                vega_wallet_path,
                "api-token",
                "generate",
                "--home=" + tmp_vega_home,
                "--tokens-passphrase-file=" + tmp_vega_home + "/passphrase-file",
                "--wallet-passphrase-file=" + tmp_vega_home + "/passphrase-file",
                "--wallet-name=" + DEFAULT_WALLET_NAME,
                "--description=" + DEFAULT_WALLET_NAME,
            ],
            capture_output=True,
        )

        wallet_args = [
            vega_wallet_path,
            "service",
            "run",
            "--network",
            "local",
            "--home=" + tmp_vega_home,
            "--automatic-consent",
            "--load-tokens",
            "--tokens-passphrase-file=" + tmp_vega_home + "/passphrase-file",
        ]

        vegaWalletProcess = _popen_process(
            wallet_args,
            dir_root=tmp_vega_dir,
            log_name="vegawallet",
        )
        processes["wallet"] = vegaWalletProcess

    if run_with_console:
        env_copy = os.environ.copy()
        env_copy.update(
            {
                "NX_VEGA_URL": (
                    f"http://localhost:{port_config[Ports.DATA_NODE_REST]}/graphql"
                ),
                "NX_VEGA_WALLET_URL": f"http://localhost:{port_config[Ports.WALLET]}",
                "NX_VEGA_ENV": "CUSTOM",
                "NX_PORT": f"{port_config[Ports.CONSOLE]}",
                "NODE_ENV": "development",
                "NX_VEGA_NETWORKS": "{}",
            }
        )
        console_process = _popen_process(
            [
                "yarn",
                "--cwd",
                vega_console_path,
                "nx",
                "serve",
                "-o",
                "trading",
                "--port",
                f"{port_config[Ports.CONSOLE]}",
            ],
            dir_root=tmp_vega_dir,
            log_name="console",
            env=env_copy,
        )
        processes["console"] = console_process

    # Send process pid values for resource monitoring
    child_conn.send({name: process.pid for name, process in processes.items()})

    # According to https://docs.oracle.com/cd/E19455-01/806-5257/gen-75415/index.html
    # There is no guarantee that signal will be catch by this thread. Usually the
    # parent process catches the signal and removes it from the list of pending
    # signals, this leave us with memory leak where we have orphaned vega processes
    # and the docker containers. Below is hack to maximize chance by catching the
    # signal.
    # We call signal.signal method as a workaround to move this thread on top of
    # the catch stack, then sigwait waits until singal is trapped.
    # As last resort We catches the `SIGCHLD`  in case the parent process exited
    # and this is the orphan now.
    # But to provide 100% guarantee this should be implemented in another way:
    #   - Signal should be trapped in the main process, and this should be synced
    #     the shared memory
    #   - or this entire process manager should be incorporated in the VegaServiceNull
    #     and containers/processes should be removed as inline call in the __exit__
    #
    #
    # Important assumption is that this signal can be caught multiple times as well
    def sighandler(signal, frame):
        if signal is None:
            logging.debug("VegaServiceNull exited normally")
        else:
            logging.debug(f"VegaServiceNull exited after trap the {signal} signal")

        logger.debug("Received signal from parent process")
        if use_docker_postgres:

            def kill_docker_container() -> None:
                try:
                    data_node_container.kill()
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        logger.debug(
                            f"Container {data_node_container.name} has been already"
                            " killed"
                        )
                        return
                    else:
                        raise e

            logger.debug(f"Stopping container {data_node_container.name}")
            retry(10, 1.0, kill_docker_container)

            removed = False
            for _ in range(10):
                try:
                    logging.info(f"Removing volume {data_node_docker_volume.name}")
                    data_node_docker_volume.remove(force=True)
                    removed = True
                    break
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        removed = True
                        logger.debug(
                            f"Data node volume {data_node_docker_volume.name} has been"
                            " already killed"
                        )
                        break
                    else:
                        time.sleep(1)
                except docker.errors.APIError:
                    time.sleep(1)
            if not removed:
                logging.exception(
                    "Docker volume failed to cleanup, will require manual cleaning"
                )

        logger.debug("Starting termination for processes")
        for name, process in processes.items():
            logger.debug(f"Terminating process {name}(pid: {process.pid})")
            process.terminate()

        for name, process in processes.items():
            attempts = 0
            while process.poll() is None:
                logger.debug(f"Process {name} still not terminated")
                time.sleep(1)
                attempts += 1
                if attempts > 60:
                    logger.warning(
                        "Gracefully terminating process timed-out. Killing process"
                        f" {name}."
                    )
                    process.kill()
            logger.debug(f"Process {name} stopped with {process.poll()}")
            if process.poll() == 0:
                logger.debug(f"Process {name} terminated.")
            if process.poll() == -9:
                logger.debug(f"Process {name} killed.")

        if not retain_log_files and os.path.exists(tmp_vega_dir):
            shutil.rmtree(tmp_vega_dir)

    # The below lines are workaround to put the signal listeners on top of the stack, so this process can handle it.
    signal.signal(signal.SIGINT, lambda _s, _h: None)
    signal.signal(signal.SIGTERM, lambda _s, _h: None)

    # The process had previously created one or more child processes with the fork() function.
    # One or more of these processes has since died.
    signal.signal(signal.SIGCHLD, sighandler)
    signal.sigwait(
        [
            signal.SIGKILL,  # The process was explicitly killed by somebody wielding the kill program.
            signal.SIGTERM,  # The process was explicitly killed by somebody wielding the terminate program.
        ]
    )
    sighandler(None, None)


class VegaServiceNull(VegaService):
    PORT_TO_FIELD_MAP = {
        Ports.WALLET: "wallet_port",
        Ports.DATA_NODE_GRPC: "data_node_grpc_port",
        Ports.DATA_NODE_REST: "data_node_rest_port",
        Ports.DATA_NODE_POSTGRES: "data_node_postgres_port",
        Ports.FAUCET: "faucet_port",
        Ports.VEGA_NODE: "vega_node_port",
        Ports.CORE_GRPC: "vega_node_grpc_port",
        Ports.CORE_REST: "vega_node_rest_port",
        Ports.CONSOLE: "console_port",
    }

    def __init__(
        self,
        vega_path: Optional[str] = None,
        data_node_path: Optional[str] = None,
        vega_wallet_path: Optional[str] = None,
        vega_console_path: Optional[str] = None,
        start_immediately: bool = False,
        run_with_console: bool = False,
        run_wallet_with_token_dapp: bool = False,
        port_config: Optional[Dict[Ports, int]] = None,
        warn_on_raw_data_access: bool = True,
        transactions_per_block: int = 1,
        seconds_per_block: int = 1,
        use_full_vega_wallet: bool = False,
        retain_log_files: bool = False,
        launch_graphql: bool = False,
        store_transactions: bool = True,
        replay_from_path: Optional[str] = None,
        listen_for_high_volume_stream_updates: bool = False,
        check_for_binaries: bool = False,
    ):
        super().__init__(
            can_control_time=True,
            warn_on_raw_data_access=warn_on_raw_data_access,
            seconds_per_block=seconds_per_block,
            listen_for_high_volume_stream_updates=listen_for_high_volume_stream_updates,
        )
        self.retain_log_files = retain_log_files

        self._using_all_custom_paths = all(
            [x is not None for x in [vega_path, data_node_path, vega_wallet_path]]
        )
        self.vega_path = vega_path or path.join(vega_bin_path, "vega")
        self.data_node_path = data_node_path or path.join(vega_bin_path, "data-node")
        self.vega_wallet_path = vega_wallet_path or path.join(
            vega_bin_path, "vegawallet"
        )
        self.vega_console_path = vega_console_path or path.join(
            vega_bin_path, "console"
        )
        self.proc = None
        self.run_with_console = run_with_console
        self.run_wallet_with_token_dapp = run_wallet_with_token_dapp

        self.transactions_per_block = transactions_per_block
        self.seconds_per_block = seconds_per_block

        self._wallet = None
        self._use_full_vega_wallet = use_full_vega_wallet
        self.store_transactions = store_transactions

        self.log_dir = tempfile.mkdtemp(prefix="vega-sim-")

        self.launch_graphql = launch_graphql
        self.replay_from_path = replay_from_path
        self.check_for_binaries = check_for_binaries

        self.stopped = False

        self._assign_ports(port_config)

        if start_immediately:
            self.start()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    def wait_fn(self, wait_multiple: float = 1) -> None:
        self.wait_for_core_catchup()
        self.forward(f"{int(wait_multiple * self.seconds_per_block)}s")
        self.wait_for_core_catchup()

    @property
    def wallet(self) -> Wallet:
        if self._wallet is None:
            if self._use_full_vega_wallet:
                self._wallet = VegaWallet(
                    self.wallet_url,
                    wallet_path=self.vega_wallet_path,
                    vega_home_dir=path.join(self.log_dir, "vegahome"),
                    passphrase_file_path=path.join(
                        self.log_dir, "vegahome", "passphrase-file"
                    ),
                )
            else:
                self._wallet = SlimWallet(
                    self.core_client,
                    full_wallet=None,
                    log_dir=self.log_dir,
                )
        return self._wallet

    def _check_started(self) -> None:
        if self.proc is None:
            raise ServiceNotStartedError("NullChain Vega accessed without starting")

    def _generate_port_config(self) -> Dict[Ports, int]:
        return {
            Ports.WALLET: self.wallet_port,
            Ports.DATA_NODE_GRPC: self.data_node_grpc_port,
            Ports.DATA_NODE_REST: self.data_node_rest_port,
            Ports.DATA_NODE_POSTGRES: self.data_node_postgres_port,
            Ports.FAUCET: self.faucet_port,
            Ports.VEGA_NODE: self.vega_node_port,
            Ports.CORE_GRPC: self.vega_node_grpc_port,
            Ports.CORE_REST: self.vega_node_rest_port,
            Ports.CONSOLE: self.console_port,
        }

    # set ports from port_config or alternatively find a free port
    # to use
    def _assign_ports(self, port_config: Optional[Dict[Ports, int]]):
        self.wallet_port = 0
        self.data_node_rest_port = 0
        self.data_node_grpc_port = 0
        self.data_node_postgres_port = 0
        self.faucet_port = 0
        self.vega_node_port = 0
        self.vega_node_grpc_port = 0
        self.vega_node_rest_port = 0
        self.console_port = 0

        for key, name in self.PORT_TO_FIELD_MAP.items():
            if port_config is not None and key in port_config:
                setattr(self, name, port_config[key])
            else:
                curr_ports = set(
                    [getattr(self, port) for port in self.PORT_TO_FIELD_MAP.values()]
                )
                setattr(self, name, find_free_port(curr_ports))

    def start(self, block_on_startup: bool = True) -> None:
        if self.check_for_binaries and not self._using_all_custom_paths:
            download_binaries()
        parent_conn, child_conn = multiprocessing.Pipe()
        ctx = multiprocessing.get_context()
        port_config = self._generate_port_config()
        self.proc = ctx.Process(
            target=manage_vega_processes,
            kwargs={
                "child_conn": child_conn,
                "vega_path": self.vega_path,
                "data_node_path": self.data_node_path,
                "vega_wallet_path": self.vega_wallet_path,
                "vega_console_path": self.vega_console_path,
                "run_with_console": self.run_with_console,
                "port_config": port_config,
                "transactions_per_block": self.transactions_per_block,
                "block_duration": f"{int(self.seconds_per_block)}s",
                "run_wallet": self._use_full_vega_wallet,
                "retain_log_files": self.retain_log_files,
                "log_dir": self.log_dir,
                "store_transactions": self.store_transactions,
                "replay_from_path": self.replay_from_path,
            },
        )
        self.proc.start()

        if self.run_with_console:
            logger.info(
                "Vega Running. Console launched at"
                f" http://localhost:{self.console_port}"
            )

        if block_on_startup:
            # Wait for startup
            started = False
            for _ in range(500):
                try:
                    channel = grpc.insecure_channel(
                        self.data_node_grpc_url,
                        options=(
                            ("grpc.enable_http_proxy", 0),
                            ("grpc.max_send_message_length", 1024 * 1024 * 20),
                            ("grpc.max_receive_message_length", 1024 * 1024 * 20),
                        ),
                    )
                    grpc.channel_ready_future(channel).result(timeout=5)
                    trading_data_client = vac.VegaTradingDataClientV2(
                        self.data_node_grpc_url,
                        channel=channel,
                    )
                    gov.get_blockchain_time(trading_data_client)

                    requests.get(
                        f"http://localhost:{self.data_node_rest_port}/time"
                    ).raise_for_status()
                    requests.get(
                        f"http://localhost:{self.vega_node_rest_port}/blockchain/height"
                    ).raise_for_status()
                    if self._use_full_vega_wallet:
                        requests.get(
                            f"http://localhost:{self.wallet_port}/api/v2/health"
                        ).raise_for_status()

                    started = True
                    break
                except (
                    MaxRetryError,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.HTTPError,
                    grpc.RpcError,
                    grpc.FutureTimeoutError,
                ):
                    time.sleep(0.1)
            if not started:
                self.stop()
                raise VegaStartupTimeoutError(
                    "Timed out waiting for Vega simulator to start up"
                )

            self.process_pids = parent_conn.recv()

        # Initialise the data-cache
        self.data_cache

        if self.run_with_console:
            webbrowser.open(f"http://localhost:{port_config[Ports.CONSOLE]}/", new=2)

        if self.launch_graphql:
            webbrowser.open(
                f"http://localhost:{port_config[Ports.DATA_NODE_REST]}/graphql", new=2
            )

    # Class internal as at some point the host may vary as well as the port
    @staticmethod
    def _build_url(port: int, prefix: str = "http://"):
        return f"{prefix}localhost:{port}"

    def stop(self) -> None:
        logging.debug("Calling stop for veganullchain")
        if self.stopped:
            return
        self.stopped = True
        self.core_client.stop()
        self.core_state_client.stop()
        self.trading_data_client_v2.stop()
        if self.proc is None:
            logger.info("Stop called but nothing to stop")
        else:
            os.kill(self.proc.pid, signal.SIGTERM)
        if isinstance(self.wallet, SlimWallet):
            self.wallet.stop()
        super().stop()

    @property
    def wallet_url(self) -> str:
        return self._build_url(self.wallet_port)

    @property
    def data_node_rest_url(self) -> str:
        return self._build_url(self.data_node_rest_port)

    @property
    def data_node_grpc_url(self) -> str:
        return self._build_url(self.data_node_grpc_port, prefix="")

    @property
    def faucet_url(self) -> str:
        return self._build_url(self.faucet_port)

    @property
    def vega_node_url(self) -> str:
        return self._build_url(self.vega_node_port)

    @property
    def vega_node_grpc_url(self) -> str:
        return self._build_url(self.vega_node_grpc_port, prefix="")

    def clone(self) -> VegaServiceNull:
        """Creates a clone of the service without the handle to other processes.

        This is required as when spinning a Nullchain service out into
        separate processes we need to start the various components in the main
        thread (as daemon processes cannot spawn daemon processes), however want
        to maintain a handle to these in the child.
        """
        return VegaServiceNull(
            self.vega_path,
            self.data_node_path,
            self.vega_wallet_path,
            start_immediately=False,
            port_config=self._generate_port_config(),
            use_full_vega_wallet=self._use_full_vega_wallet,
            warn_on_raw_data_access=self.warn_on_raw_data_access,
        )



from math import log, sqrt, exp
from scipy.stats import lognorm
from typing import Optional
from deprecated import deprecated

import vega_sim.proto.vega as vega_protos


@deprecated(
    reason=(
        "This function is currently unused, so is unmaintained, but worked correctly"
        " when last used"
    )
)
def probability_of_trading(
    side: vega_protos.vega.Side,
    price: float,
    best_bid_price: float,
    best_ask_price: float,
    min_valid_price: Optional[float],
    max_valid_price: Optional[float],
    mu: float,
    tau: float,
    sigma: float,
    min_probability_of_trading: float,
) -> float:
    """Compute the probability of trading of a given order.

    Probability of trading follows the cdf of a lognormal distribution. The distribution
    is bound between the min_valid_price and the best_bid_price for buy orders, and the
    best_ask_price and max_valid_price for sell orders. All probabilities are scaled by
    a half and cannot be lower than the network parameter, min_probability_of_trading.

    Args:
        side (vega_protos.vega.Side):
            Side of the order to evaluate.
        price (float):
            Price of the order to evaluate.
        best_bid_price (float):
            Best bid price on the order book.
        best_ask_price (float):
            Best ask price on the order book.
        min_valid_price (float):
            Minimum valid price from price monitoring bounds.
        max_valid_price (float):
            Maximum valid price from price monitoring bounds.
        mu (float):
            Market parameter for the risk model.
        tau (float):
            Market parameter for the risk model.
        sigma (float):
            Market parameter for the risk model.
        min_probability_of_trading (float):
            Network parameter defining the minimum value to return.

    Returns:
        float:
            Probability of trading of the order.
    """

    if price > best_bid_price and price < best_ask_price:
        return 0.5
    elif (
        min_valid_price is None
        or max_valid_price is None
        or price < min_valid_price
        or price > max_valid_price
    ):
        return min_probability_of_trading

    if side == vega_protos.vega.SIDE_BUY:
        best_price = best_bid_price

        lower_bound = min_valid_price
        upper_bound = best_bid_price

    else:
        best_price = best_ask_price

        lower_bound = best_ask_price
        upper_bound = max_valid_price

    stdev = sigma * sqrt(tau)
    m = log(best_price) + (mu - 0.5 * sigma**2) * tau

    rv = lognorm(s=stdev, scale=exp(m))

    min = rv.cdf(lower_bound)
    max = rv.cdf(upper_bound)
    z = max - min

    if side == vega_protos.vega.SIDE_BUY:
        p = 0.5 * (rv.cdf(price) - min) / z
    else:
        p = 0.5 * (max - rv.cdf(price)) / z

    if p < min_probability_of_trading:
        return min_probability_of_trading
    else:
        return p



import os
import pickle
from collections import namedtuple
from dataclasses import dataclass
from functools import partial
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical
from tqdm import tqdm

from vega_sim.reinforcement.helpers import apply_funcs, to_torch, toggle
from vega_sim.reinforcement.networks import FFN, FFN_fix_fol_Q, Softmax
from vega_sim.reinforcement.v2.agents.learning_agent import TorchLearningAgent
from vega_sim.reinforcement.v2.agents.puppets import (
    MarketOrderAction,
    NoAction,
    Side,
    Action,
)
from vega_sim.reinforcement.v2.states import PriceStateWithFees

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Send selling/buying MOs to hit LP orders
WALLET = WalletConfig("learner", "learner")


class LearningAgentFixedVol(TorchLearningAgent):
    def __init__(
        self,
        device: str,
        logfile_pol_imp: str,
        logfile_pol_eval: str,
        logfile_pnl: str,
        discount_factor: float,
        num_levels: int,
        inventory_penalty: float = 1.0,
    ):
        super().__init__(
            device=device,
            logfile_pol_imp=logfile_pol_imp,
            logfile_pol_eval=logfile_pol_eval,
            logfile_pnl=logfile_pnl,
            discount_factor=discount_factor,
            inventory_penalty=inventory_penalty,
        )
        self.volume = 1

        # Dimensions of state and action
        self.num_levels = num_levels
        self.state_dim = 4 + 4 * self.num_levels  # from MarketState
        action_discrete_dim = 3

        # NN for Q-fun and its optimizer
        self.q_func = FFN_fix_fol_Q(state_dim=self.state_dim)
        self.optimizer_q = torch.optim.RMSprop(self.q_func.parameters(), lr=0.01)

        # NN for policy and its optimizer
        self.policy_discr = FFN(
            sizes=[self.state_dim, 4096, action_discrete_dim],
            activation=nn.ReLU,
            output_activation=Softmax,
        )  # this network decides whether to buy/sell/do nothing
        self.optimizer_pol = torch.optim.RMSprop(
            list(self.policy_discr.parameters()), lr=0.001
        )
        self.prev_state_action_reward = None

    def move_to_device(self):
        self.q_func.to(self.device)
        self.policy_discr.to(self.device)

    def move_to_cpu(self):
        self.q_func.to("cpu")
        self.policy_discr.to("cpu")

    def learning_step(self, results_dir: Optional[str] = None):
        # Policy evaluation + Policy improvement
        self.move_to_device()
        self.policy_eval(batch_size=20000, n_epochs=10)
        self.policy_improvement(batch_size=100_000, n_epochs=10)

        if results_dir is not None:
            # save in case environment chooses to crash
            self.save(results_dir)
        self.move_to_cpu()

    def _update_memory(
        self,
        state: PriceStateWithFees,
        action: Action,
        reward: float,
    ):
        """
        Ensure what is the input?
        """
        if (
            self.prev_state_action_reward is not None
            and self.prev_state_action_reward[1] is not NoAction
        ):
            prev_state, prev_action, prev_reward = self.prev_state_action_reward
            self.memory["state"].append(prev_state.to_array())
            # action_discrete = 0 if sell
            # action_discrete = 1 if buy
            # action_discrete = 2 if they do not do anything
            self.memory["action_discrete"].append([prev_action.side.value])

            self.memory["reward"].append([prev_reward])
            if state is not np.nan:
                self.memory["next_state"].append(state.to_array())
            else:
                self.memory["next_state"].append(
                    np.nan * np.ones_like(state.to_array())
                )

        self.prev_state_action_reward = (state, action, reward)

    def create_dataloader(self, batch_size):
        """
        creates dataset and dataloader for training.
        """
        to_torch_device = partial(to_torch, device=self.device)
        dataset_state = apply_funcs(
            value=self.memory["state"], funcs=(np.stack, to_torch_device)
        )
        dataset_action_discrete = apply_funcs(
            value=self.memory["action_discrete"],
            funcs=(np.array, partial(to_torch, dtype=torch.int64, device=self.device)),
        )
        dataset_reward = apply_funcs(
            value=self.memory["reward"], funcs=(np.array, to_torch_device)
        )
        dataset_next_state = apply_funcs(
            value=self.memory["next_state"], funcs=(np.stack, to_torch_device)
        )

        dataset = torch.utils.data.TensorDataset(
            *(
                dataset_state,
                dataset_action_discrete,
                dataset_reward,
                dataset_next_state,
            )
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, shuffle=True
        )
        return dataloader

    def step(self, state: PriceStateWithFees):
        self.step_num += 1
        self.latest_action = self._step(state)
        self.latest_state = state

        if state.full_balance <= 0:
            return NoAction
        if state.market_in_auction:
            return NoAction
        return self.latest_action

    def _step(self, state: PriceStateWithFees) -> Action:
        # learned policy
        state = state.to_array().reshape(1, -1)  # adding batch_dimension
        state = torch.from_numpy(state).float()  # .to(self.device)

        with torch.no_grad():
            c = self.sample_action(state=state, sim=True)
        choice = int(c.item())

        return MarketOrderAction(side=Side(choice), volume=self.volume)

    def sample_action(
        self,
        state: torch.Tensor,
        sim: bool = True,
        evaluate: bool = False,
    ):
        """
        Sample an action.

        Returns
        -------
        c is a tensor of shape (batch_size, 1) returning the sampled action from {sell, buy, do nothing}
        (i.e. c is filled with values from {0,1,2})
        """

        probs = self.policy_discr(state)
        if sim:
            m = Categorical(probs)
            c = m.sample()
        else:
            c = probs
        if evaluate:
            c = torch.max(probs, 1, keepdim=True)[1]
        return c

    def policy_eval(
        self,
        batch_size: int,
        n_epochs: int,
    ):
        toggle(self.policy_discr, to=False)
        toggle(self.q_func, to=True)

        dataloader = self.create_dataloader(batch_size=batch_size)

        pbar = tqdm(total=n_epochs)
        for epoch in range(n_epochs):
            for (
                i,
                (
                    batch_state,
                    batch_action_discrete,
                    batch_reward,
                    batch_next_state,
                ),
            ) in enumerate(dataloader):
                next_state_terminal = torch.isnan(
                    batch_next_state
                ).float()  # shape (batch_size, dim_state)
                batch_next_state[next_state_terminal.eq(True)] = batch_state[
                    next_state_terminal.eq(True)
                ]
                self.optimizer_q.zero_grad()

                pred = torch.gather(
                    self.q_func(batch_state),
                    dim=1,
                    index=batch_action_discrete,
                )

                with torch.no_grad():
                    v = self.v_func(batch_next_state)
                    target = (
                        batch_reward
                        + (1 - next_state_terminal.float().mean(1, keepdim=True))
                        * self.discount_factor
                        * v
                    )
                loss = torch.pow(pred - target, 2).mean()
                loss.backward()
                self.optimizer_q.step()
            self.losses["q"].append(loss.item())
            # logging loss
            with open(self.logfile_pol_eval, "a") as f:
                f.write(
                    "{},{:.2e},{:.2e},{:.2e}\n".format(
                        epoch + self.learningIteration * n_epochs,
                        loss.item(),
                        self.coefH_discr,
                        self.coefH_cont,
                    )
                )
            pbar.update(1)
        return 0

    def v_func(self, state):
        """
        v(x) = E[q(x,A)] = E[q(x,A)|C=0]p(C=0) + E[q(x,A)|C=1]p(C=1) + E[q(x)|C=2]p(C=2)

        if C==0 --> sell,
        if C==1 --> buy,
        if C==2 --> does nothing,
        Parameters
        ----------
        state: torch.Tensor
            State. Tensor of shape (batch_size,
        n_mc: torch.Tensor
            Number of Monte Carlo samples to calculate
        """

        q = self.q_func(state)
        c = self.sample_action(
            state, sim=False
        )  # .unravel().reshape((state.shape[0], 1))

        v = c[:, 0] * (q[:, 0] - self.coefH_discr * torch.log(c[:, 0]))
        v += c[:, 1] * (q[:, 1] - self.coefH_discr * torch.log(c[:, 1]))
        v += c[:, 2] * (q[:, 2] - self.coefH_discr * torch.log(c[:, 2]))
        self.q_func.train()
        return v

    def D_KL(self, state):
        """
        KL divergence between pi(.|x) and the unnormalised density exp(Q(x,.)),
        where pi(.|x) is a LogNormal distribution
        """
        # c = self.sample_action(state).reshape((state.shape[0], 1))

        q = self.q_func(state)
        c = self.sample_action(
            state, sim=False
        )  # .unravel().reshape((state.shape[0], 1))

        d_kl = c[:, 0] * (self.coefH_discr * torch.log(c[:, 0]) - q[:, 0])
        d_kl += c[:, 1] * (self.coefH_discr * torch.log(c[:, 1]) - q[:, 1])
        d_kl += c[:, 2] * (self.coefH_discr * torch.log(c[:, 2]) - q[:, 2])

        return d_kl

    def policy_improvement(self, batch_size: int, n_epochs: int):
        toggle(self.policy_discr, to=True)
        toggle(self.q_func, to=False)

        dataloader = self.create_dataloader(batch_size=batch_size)

        pbar = tqdm(total=n_epochs)
        for epoch in range(n_epochs):
            for i, (batch_state, _, _, _) in enumerate(dataloader):
                self.optimizer_pol.zero_grad()
                d_kl = self.D_KL(batch_state).mean()
                d_kl.backward()
                # nn.utils.clip_grad_norm_(self.policy_volume.parameters(), max_norm=1.)
                self.optimizer_pol.step()
            self.losses["d_kl"].append(d_kl.item())
            with open(self.logfile_pol_imp, "a") as f:
                f.write(
                    "{},{:.4f}\n".format(
                        epoch + n_epochs * self.learningIteration, d_kl.item()
                    )
                )
            pbar.update(1)

        # update the coefficients for the next run
        self.coefH_discr = max(self.coefH_discr * 0.99, 1e-10)
        self.coefH_cont = max(self.coefH_cont * 0.99, 1e-10)
        self.learningIteration += 1
        return 0

    def save(self, results_dir: str):
        filename = os.path.join(results_dir, "agent.pth.tar")
        d = {
            "losses": self.losses,
            "q": self.q_func.state_dict(),
            "policy_discr": self.policy_discr.state_dict(),
            "iteration": self.learningIteration,
            "coefH_discr": self.coefH_discr,
        }
        torch.save(d, filename)

        filename_for_memory = os.path.join(results_dir, "memory.pickle")
        with open(filename_for_memory, "wb") as f:
            pickle.dump(self.memory, f)

    def load(self, results_dir: str):
        filename = os.path.join(results_dir, "agent.pth.tar")
        d = torch.load(filename, map_location="cpu")
        self.q_func.load_state_dict(d["q"])
        self.policy_discr.load_state_dict(d["policy_discr"])

        self.learningIteration = d["iteration"]
        self.coefH_discr = d["coefH_discr"]

        filename_for_memory = os.path.join(results_dir, "memory.pickle")
        with open(filename_for_memory, "rb") as f:
            memory = pickle.load(f)
        self.memory = memory



from stable_baselines3 import PPO, DQN

import vega_sim.reinforcement.v2.stable_baselines.environment as env
from vega_sim.reinforcement.v2.states import PriceStateWithFees, PositionOnly

if __name__ == "__main__":
    e = env.SingleAgentVegaEnv(
        action_type=env.ActionType.AT_TOUCH_ONE_SIDE,
        steps_per_trading_session=200,
        reward_type=env.Reward.PNL,
        terminal_reward_type=env.Reward.SQ_INVENTORY_PENALTY,
        state_type=PriceStateWithFees,
    )
    model = PPO(
        "MlpPolicy",
        e,
        verbose=1,
        tensorboard_log="./ppo_tensorboard/",
        n_steps=600,
        batch_size=50,
    ).learn(total_timesteps=1_000_000)



import os

import logging
import argparse

from vega_sim.null_service import VegaServiceNull

from vega_sim.scenario.constants import Network
from vega_sim.scenario.fuzzed_markets.scenario import FuzzingScenario

from vega_sim.tools.scenario_plots import (
    fuzz_plots,
    plot_run_outputs,
    account_plots,
    plot_price_monitoring,
    reward_plots,
)

from matplotlib import pyplot as plt


def _run(
    steps: int = 2880,
    console: bool = False,
    output: bool = False,
    output_dir: str = "fuzz_plots",
):
    scenario = FuzzingScenario(
        num_steps=steps,
        step_length_seconds=30,
        block_length_seconds=1,
        transactions_per_block=4096,
        output=output,
    )

    with VegaServiceNull(
        warn_on_raw_data_access=False,
        seconds_per_block=scenario.block_length_seconds,
        transactions_per_block=scenario.transactions_per_block,
        retain_log_files=True,
        use_full_vega_wallet=False,
        run_with_console=console,
    ) as vega:
        scenario.run_iteration(
            vega=vega,
            network=Network.NULLCHAIN,
            output_data=output,
            log_every_n_steps=100,
        )

    if output:
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        fuzz_figs = plot_price_monitoring()
        for key, fig in fuzz_figs.items():
            fig.savefig(f"{output_dir}/monitoring-{key}.jpg")
            plt.close(fig)

        fuzz_figs = fuzz_plots()
        for key, fig in fuzz_figs.items():
            fig.savefig(f"{output_dir}/fuzz-{key}.jpg")
            plt.close(fig)

        trading_figs = plot_run_outputs()
        for key, fig in trading_figs.items():
            fig.savefig(f"{output_dir}/trading-{key}.jpg")
            plt.close(fig)

        reward_fig = reward_plots()
        reward_fig.savefig(f"{output_dir}/rewards.jpg")
        plt.close(fig)

        account_fig = account_plots()
        account_fig.savefig(f"{output_dir}/accounts.jpg")
        plt.close(account_fig)


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s",
        "--steps",
        default=2 * 60 * 12,
        type=int,
    )
    parser.add_argument("--console", action="store_true")
    args = parser.parse_args()

    _run(steps=args.steps, console=args.console, output=True)



"""wallet.py

Module contains agent information variables for each party in the devops scenario. An 
agent variable contains the wallet name and key name for an agent.

When deploying a devops scenario to an existing network; A wallet with the following key
names must exist locally.

    • market_creator
    • market_settler
    • market_maker
    • auction_trader_a
    • auction_trader_b
    • random_trader_a
    • random_trader_b
    • random_trader_c
    • sensitive_trader_a
    • sensitive_trader_b
    • sensitive_trader_c

The wallet name and wallet passphrase can be modified with the environment variables
VEGA_USER_WALLET_NAME and VEGA_USER_WALLETS_PASS.

"""

import os
import dotenv

from dataclasses import dataclass

DEFAULT_USER_WALLET_NAME = "vega-market-sim"
DEFAULT_USER_WALLET_PASS = "passphrase"


# Set the default wallet_name and wallet_pass in this dataclass
@dataclass
class Agent:
    dotenv.load_dotenv()
    wallet_name: str = os.environ.get("VEGA_USER_WALLET_NAME", DEFAULT_USER_WALLET_NAME)
    wallet_pass: str = os.environ.get("VEGA_USER_WALLET_PASS", DEFAULT_USER_WALLET_PASS)
    key_name: str = None


# Market proper creates and approves a market creation proposal
MARKET_CREATOR_AGENT = Agent(key_name="market_creator")

# Market terminator creates and approves a market settle proposal
MARKET_SETTLER_AGENT = Agent(key_name="market_settler")

# Market maker makes a market by providing liquidity and an order book
MARKET_MAKER_AGENT = Agent(key_name="market_maker")

# Auction traders provide trades in auction trading modes
AUCTION_TRADER_AGENTS = [
    Agent(key_name="auction_trader_a"),
    Agent(key_name="auction_trader_b"),
]

# Random traders provide trades in continuous trading modes
RANDOM_TRADER_AGENTS = [
    Agent(key_name="random_trader_a"),
    Agent(key_name="random_trader_b"),
    Agent(key_name="random_trader_c"),
]

# Sensitive traders exploit over-exposed positions in continuous trading modes
SENSITIVE_TRADER_AGENTS = [
    Agent(key_name="sensitive_trader_a"),
    Agent(key_name="sensitive_trader_b"),
    Agent(key_name="sensitive_trader_c"),
]



import os

import torch
from torch.utils.tensorboard import SummaryWriter
from pettingzoo.utils.conversions import parallel_to_aec_wrapper
from tianshou.data import Collector, VectorReplayBuffer
from tianshou.env import DummyVectorEnv, PettingZooEnv
from tianshou.policy import BasePolicy, DQNPolicy, MultiAgentPolicyManager, RandomPolicy
from tianshou.trainer import offpolicy_trainer
from tianshou.utils import TensorboardLogger
from tianshou.utils.net.common import Net

import vega_sim.reinforcement.v2.pettingzoo.environment as pz


def get_agents(env: pz.MultiAgentVegaEnv):
    agents = []
    for agent in env.agents:
        model = Net(
            action_shape=env.action_spaces[agent].n,
            state_shape=(
                env.observation_spaces[agent].shape[0]
                if len(env.observation_spaces[agent].shape) > 0
                else 1
            ),
            hidden_sizes=[10],
        )
        optim = torch.optim.Adam(model.parameters())

        agents.append(DQNPolicy(model, optim))

    return agents


if __name__ == "__main__":
    env = None
    # Step 1: Load the PettingZoo environment
    try:
        agent_configs = [
            pz.AgentConfig(
                action_type=pz.ActionType.AT_TOUCH_ONE_SIDE,
                state_type=pz.PriceStateWithFees,
                # reward_type=pz.Reward.PNL,
                # terminal_reward_type=pz.Reward.SQ_INVENTORY_PENALTY,
            ),
            pz.AgentConfig(
                action_type=pz.ActionType.AT_TOUCH_ONE_SIDE,
                state_type=pz.PriceStateWithFees,
                # reward_type=pz.Reward.PNL,
                # terminal_reward_type=pz.Reward.SQ_INVENTORY_PENALTY,
            ),
        ]
        base_env = pz.MultiAgentVegaEnv(
            agents=agent_configs,
            steps_per_trading_session=200,
            unified_reward=pz.Reward.UNIFIED_PNL,
        )

        # Step 2: Wrap the environment for Tianshou interfacing
        # env.reset()
        env = PettingZooEnv(parallel_to_aec_wrapper(base_env))
        # Step 3: Define policies for each agent
        agents = get_agents(base_env)
        policy = MultiAgentPolicyManager(agents, env)

        # Step 4: Convert the env to vector format
        train_env = DummyVectorEnv([lambda: env])
        test_env = DummyVectorEnv([lambda: env])

        # Step 5: Construct the Collector, which interfaces the policies with the vectorised environment
        train_collector = Collector(
            policy, train_env, VectorReplayBuffer(100, 1), exploration_noise=True
        )
        test_collector = Collector(policy, test_env, exploration_noise=True)

        # Step 6: Execute the environment with the agents playing for 1 episode, and render a frame every 0.1 seconds
        train_collector.collect(n_episode=10)

        def save_best_fn(policy):
            pass

        # def stop_fn(mean_rewards):
        #     return mean_rewards >= 100

        def train_fn(epoch, env_step):
            for p in policy.policies.values():
                p.set_eps(0.1)

        def test_fn(epoch, env_step):
            for p in policy.policies.values():
                p.set_eps(0.05)

        def reward_fn(rewards):
            return rewards.sum(axis=1)

        # ======== tensorboard logging setup =========
        log_path = os.path.join(
            "./",
            "tianshou",
            "dqn",
        )
        writer = SummaryWriter(log_path)
        logger = TensorboardLogger(writer)

        # trainer
        result = offpolicy_trainer(
            policy,
            train_collector,
            test_collector,
            max_epoch=300,
            step_per_epoch=600,
            step_per_collect=50,
            episode_per_test=3,
            batch_size=64,
            train_fn=train_fn,
            test_fn=test_fn,
            reward_metric=reward_fn,
            # stop_fn=stop_fn,
            save_best_fn=save_best_fn,
            update_per_step=0.1,
            logger=logger,
            test_in_train=False,
        )
    finally:
        if env is not None:
            env.close()



"""run_scenario.py

Script used to run vega-market-sim scenarios configure for devops.

Flags:

    -s, --scenario          controls which scenario to test/deploy
    -n, --network           controls which network the scenario should run on
    -m, --market            controls which market the market should run on

    -d, --debug             controls whether to log debug messages
    -c, --console           whether to launch a console
    -g, --graphql           whether to launch a graphql playground 
    -p, --pause             whether to pause a the end of a simulation
    

Examples:

    Test the ETHUSD scenario with a simulation on a nullchain network.

    $ python -m vega_sim.vega_sim.devops.run -s ETHUSD -n NULLCHAIN -d -c -p


"""

import argparse
import logging

from vega_sim.scenario.constants import Network

from vega_sim.null_service import VegaServiceNull
from vega_sim.network_service import VegaServiceNetwork

from vega_sim.devops.scenario import DevOpsScenario
from vega_sim.devops.registry import SCENARIOS


def main():
    parser = argparse.ArgumentParser()

    # Simulation / deployment arguments
    parser.add_argument(
        "-n",
        "--network",
        choices=[network.name for network in Network],
        default=Network.FAIRGROUND.name,
    )
    parser.add_argument("-s", "--scenario", default=None, type=str)
    parser.add_argument("-m", "--market_name", default=None, type=str)
    parser.add_argument("-l", "--step_length_seconds", default=10, type=int)

    # Developer arguments
    parser.add_argument("-c", "--console", action="store_true")
    parser.add_argument("-p", "--pause", action="store_true")
    parser.add_argument("-d", "--debug", action="store_true")

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    scenario: DevOpsScenario = SCENARIOS[args.scenario]()

    scenario.market_name = args.market_name
    scenario.step_length_seconds = args.step_length_seconds

    if Network[args.network] == Network.NULLCHAIN:
        with VegaServiceNull(
            seconds_per_block=1,
            transactions_per_block=1000,
            retain_log_files=True,
            use_full_vega_wallet=True,
            warn_on_raw_data_access=False,
            run_with_console=args.console,
        ) as vega:
            scenario.run_iteration(
                vega=vega,
                network=Network[args.network],
                pause_at_completion=args.pause,
            )

    else:
        with VegaServiceNetwork(
            network=Network[args.network],
            run_with_console=args.console,
        ) as vega:
            scenario.run_iteration(
                vega=vega,
                network=Network[args.network],
                pause_at_completion=args.pause,
                raise_datanode_errors=False,
                raise_step_errors=False,
                run_with_snitch=False,
            )


if __name__ == "__main__":
    main()



import argparse
import logging
import os
import platform
import shutil
import tempfile
import zipfile

import requests

import vega_sim

logger = logging.getLogger(__name__)

URL_BASE = "https://github.com/vegaprotocol/vega/releases/download/v{version}/"
URL_LATEST_BASE = "https://github.com/vegaprotocol/vega/releases/latest/download/"
DATA_NODE = "data-node-{platform}-{chipset}64.zip"
VEGA_CORE = "vega-{platform}-{chipset}64.zip"
VEGAWALLET = "vegawallet-{platform}-{chipset}64.zip"

FILES = [DATA_NODE, VEGA_CORE, VEGAWALLET]

BIN_NAMES = {
    DATA_NODE: "data-node",
    VEGA_CORE: "vega",
    VEGAWALLET: "vegawallet",
}


def download_binaries(force: bool = False, latest: bool = False):
    platf = platform.system().lower()

    # We have no Windows specific builds, so people have to use WSL to run
    platf = "linux" if platf == "windows" else platf

    chipset = "arm" if "arm" in platform.machine() else "amd"

    if force and os.path.exists(vega_sim.vega_bin_path):
        shutil.rmtree(vega_sim.vega_bin_path)

    with tempfile.TemporaryDirectory() as dir:
        if not os.path.exists(vega_sim.vega_bin_path):
            os.mkdir(vega_sim.vega_bin_path)
        for remote_file in FILES:
            file_name = remote_file.format(platform=platf, chipset=chipset)
            if (
                not os.path.exists(
                    os.path.join(vega_sim.vega_bin_path, BIN_NAMES[remote_file])
                )
            ) or force:
                logger.info(f"Downloading {file_name}")

                url_base = (
                    URL_BASE.format(version=vega_sim.VEGA_VERSION)
                    if not latest
                    else URL_LATEST_BASE
                )
                url = url_base + file_name
                res = requests.get(url)
                file_path = os.path.join(dir, f"{file_name}")

                with open(file_path, "wb") as f:
                    f.write(res.content)
                with zipfile.ZipFile(file_path, "r") as zip_ref:
                    zip_ref.extractall(vega_sim.vega_bin_path)

        for bin_file in os.listdir(vega_sim.vega_bin_path):
            os.chmod(os.path.join(vega_sim.vega_bin_path, bin_file), 0o775)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("--latest", action="store_true")
    args = parser.parse_args()

    download_binaries(force=args.force, latest=args.latest)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/data_source.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from .data.v1 import data_pb2 as vega_dot_data_dot_v1_dot_data__pb2
from .data.v1 import spec_pb2 as vega_dot_data_dot_v1_dot_spec__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16vega/data_source.proto\x12\x04vega\x1a\x1cgoogle/protobuf/struct.proto\x1a\x17vega/data/v1/data.proto\x1a\x17vega/data/v1/spec.proto"\xa9\x01\n\x14\x44\x61taSourceDefinition\x12@\n\x08internal\x18\x01 \x01(\x0b\x32".vega.DataSourceDefinitionInternalH\x00R\x08internal\x12@\n\x08\x65xternal\x18\x02 \x01(\x0b\x32".vega.DataSourceDefinitionExternalH\x00R\x08\x65xternalB\r\n\x0bsource_type"Z\n\x1f\x44\x61taSourceSpecConfigurationTime\x12\x37\n\nconditions\x18\x01 \x03(\x0b\x32\x17.vega.data.v1.ConditionR\nconditions"j\n\x1c\x44\x61taSourceDefinitionInternal\x12;\n\x04time\x18\x01 \x01(\x0b\x32%.vega.DataSourceSpecConfigurationTimeH\x00R\x04timeB\r\n\x0bsource_type"\x9a\x01\n\x1c\x44\x61taSourceDefinitionExternal\x12;\n\x06oracle\x18\x01 \x01(\x0b\x32!.vega.DataSourceSpecConfigurationH\x00R\x06oracle\x12.\n\x08\x65th_call\x18\x02 \x01(\x0b\x32\x11.vega.EthCallSpecH\x00R\x07\x65thCallB\r\n\x0bsource_type"}\n\x1b\x44\x61taSourceSpecConfiguration\x12.\n\x07signers\x18\x01 \x03(\x0b\x32\x14.vega.data.v1.SignerR\x07signers\x12.\n\x07\x66ilters\x18\x02 \x03(\x0b\x32\x14.vega.data.v1.FilterR\x07\x66ilters"\xc9\x01\n\x0b\x45thCallSpec\x12\x18\n\x07\x61\x64\x64ress\x18\x01 \x01(\tR\x07\x61\x64\x64ress\x12,\n\x03\x61\x62i\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValueR\x03\x61\x62i\x12\x16\n\x06method\x18\x03 \x01(\tR\x06method\x12*\n\x04\x61rgs\x18\x04 \x03(\x0b\x32\x16.google.protobuf.ValueR\x04\x61rgs\x12.\n\x07trigger\x18\x05 \x01(\x0b\x32\x14.vega.EthCallTriggerR\x07trigger"V\n\x0e\x45thCallTrigger\x12\x39\n\x0ctime_trigger\x18\x01 \x01(\x0b\x32\x14.vega.EthTimeTriggerH\x00R\x0btimeTriggerB\t\n\x07trigger"\x85\x01\n\x0e\x45thTimeTrigger\x12\x1d\n\x07initial\x18\x01 \x01(\x04H\x00R\x07initial\x88\x01\x01\x12\x19\n\x05\x65very\x18\x02 \x01(\x04H\x01R\x05\x65very\x88\x01\x01\x12\x19\n\x05until\x18\x03 \x01(\x04H\x02R\x05until\x88\x01\x01\x42\n\n\x08_initialB\x08\n\x06_everyB\x08\n\x06_until"\x90\x02\n\x0e\x44\x61taSourceSpec\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1d\n\ncreated_at\x18\x02 \x01(\x03R\tcreatedAt\x12\x1d\n\nupdated_at\x18\x03 \x01(\x03R\tupdatedAt\x12.\n\x04\x64\x61ta\x18\x04 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR\x04\x64\x61ta\x12\x33\n\x06status\x18\x05 \x01(\x0e\x32\x1b.vega.DataSourceSpec.StatusR\x06status"K\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x11\n\rSTATUS_ACTIVE\x10\x01\x12\x16\n\x12STATUS_DEACTIVATED\x10\x02"B\n\x16\x45xternalDataSourceSpec\x12(\n\x04spec\x18\x01 \x01(\x0b\x32\x14.vega.DataSourceSpecR\x04specB\'Z%code.vegaprotocol.io/vega/protos/vegab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "vega.data_source_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z%code.vegaprotocol.io/vega/protos/vega"
    _globals["_DATASOURCEDEFINITION"]._serialized_start = 113
    _globals["_DATASOURCEDEFINITION"]._serialized_end = 282
    _globals["_DATASOURCESPECCONFIGURATIONTIME"]._serialized_start = 284
    _globals["_DATASOURCESPECCONFIGURATIONTIME"]._serialized_end = 374
    _globals["_DATASOURCEDEFINITIONINTERNAL"]._serialized_start = 376
    _globals["_DATASOURCEDEFINITIONINTERNAL"]._serialized_end = 482
    _globals["_DATASOURCEDEFINITIONEXTERNAL"]._serialized_start = 485
    _globals["_DATASOURCEDEFINITIONEXTERNAL"]._serialized_end = 639
    _globals["_DATASOURCESPECCONFIGURATION"]._serialized_start = 641
    _globals["_DATASOURCESPECCONFIGURATION"]._serialized_end = 766
    _globals["_ETHCALLSPEC"]._serialized_start = 769
    _globals["_ETHCALLSPEC"]._serialized_end = 970
    _globals["_ETHCALLTRIGGER"]._serialized_start = 972
    _globals["_ETHCALLTRIGGER"]._serialized_end = 1058
    _globals["_ETHTIMETRIGGER"]._serialized_start = 1061
    _globals["_ETHTIMETRIGGER"]._serialized_end = 1194
    _globals["_DATASOURCESPEC"]._serialized_start = 1197
    _globals["_DATASOURCESPEC"]._serialized_end = 1469
    _globals["_DATASOURCESPEC_STATUS"]._serialized_start = 1394
    _globals["_DATASOURCESPEC_STATUS"]._serialized_end = 1469
    _globals["_EXTERNALDATASOURCESPEC"]._serialized_start = 1471
    _globals["_EXTERNALDATASOURCESPEC"]._serialized_end = 1537
# @@protoc_insertion_point(module_scope)
